/* exs_extra.c */

/*
 * Copyright 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228 and Grant No. DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _POSIX_C_SOURCE 200112L
#define _XOPEN_SOURCE 600

#include <assert.h>
#include <errno.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>

#include <exs.h>


/* Waits for events, and dispatches the events to the handler for the given
 * type.  Continues waiting in a loop until a handler returns less than 0, or
 * the timeout is exceeded.
 *
 * Returns 0 if errno was not set by the last handler, or the negative errno
 * value if the last handler set errno. */
int
dispatch_exs_events(exs_qhandle_t qh, int max_count,
					int (* const handlers[])(exs_event_t *),
					const struct timeval *timeout)
{
	exs_event_t events[max_count];
	exs_event_t *ev;
	int i, n, ret;

	assert(qh != EXS_QHANDLE_INVALID);
	assert(max_count > 0);

	while (1) {
		if ((n = exs_qdequeue(qh, events, max_count, timeout)) < 0) {
			return -errno;
		} else if (n == 0) {
			return 0;
		}

		for (i = 0; i < n; i++) {
			ev = events + i;
			assert(ev->exs_evt_type >= EXS_EVT_POLL
					&& ev->exs_evt_type <= EXS_EVT_SHUTDOWN);
			errno = 0;
			if ((ret = (*handlers[ev->exs_evt_type])(ev)) < 0) {
				if (errno != 0) {
					fprintf(stderr, "event %s returned %d: %s\n",
						exs_evt_type_to_string(ev->exs_evt_type),
						ret, strerror(errno));
				}
				return -errno;
			}
		}
	}

	/* Should never get here */
	assert(0);
	return -EINVAL;
}	/* dispatch_exs_events */


static int
our_try_connect_new(const struct addrinfo *const aptr,
		int flags, struct timeval *timeout,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle)
{
	int err;
	int fd;

	if ((fd = exs_socket(aptr->ai_family, aptr->ai_socktype, 0)) < 0) {
		goto out;
	}

	if ((exs_connect(fd, aptr->ai_addr, aptr->ai_addrlen, flags, timeout,
						qhandle, ahandle) < 0)) {
		err = errno;
		exs_blocking_close(fd);
		errno = err;
		fd = -1;
	}

out:
	return fd;
}	/* our_try_connect_new */


/* Starts a connection request to the given host and port and returns the fd of
 * the new socket.  The connection request will be satisfied asynchronously, and
 * the completion event will be delivered on the given qhandle.
 *
 * This is just a simple wrapper around calling both exs_socket() and
 * exs_connect() in a row.  Note that most socket options would have to be set
 * in our_try_connect_new(). */
int
exs_connect_new(const char *const host, const char *const port,
		int family, int socktype, int flags, struct timeval *timeout,
		exs_qhandle_t qhandle, exs_ahandle_t ahandle)
{
	struct addrinfo *info, *aptr;
	struct addrinfo hints;
	int errcode;
	int fd;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = family;
	hints.ai_socktype = socktype;
	hints.ai_protocol = 0;
	hints.ai_flags = AI_ADDRCONFIG | AI_V4MAPPED;

	fd = -1;
	if ((errcode = getaddrinfo(host, port, &hints, &info)) != 0) {
		/* FIXME: better way of reporting this error */
		errno = ENETUNREACH;
		goto err1;
	}

	for (aptr = info; aptr != NULL; aptr = aptr->ai_next) {
		fd = our_try_connect_new(aptr, flags, timeout,
							qhandle, ahandle);
		if (fd >= 0) {
			break;
		}
	}

	freeaddrinfo(info);
err1:
	return fd;
}	/* exs_connect_new */


/* Returns an array of sockets which have been bound to some local address.
 * The socktype to use must be specified, but all other fields are automagically
 * determined.  If flags contains SOCKET_LISTEN, also listen on the socket. */
int
our_bind_all(const char *hostname, const char *service,
		int family, int socktype,
		int **fds, int *nfds)
{
	struct addrinfo *info, *aptr;
	struct addrinfo hints;
	int *fdlist;
	int errcode;
	int next_fd;
	int max_count;
	int count;
	int retval;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = family;
	hints.ai_socktype = socktype;
	hints.ai_protocol = 0;
	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG | AI_V4MAPPED;

	if ((errcode = getaddrinfo(hostname, service, &hints, &info)) != 0) {
#if 0
		fprintf(stderr, "Could not resolve port %s: %s\n", service,
				gai_strerror(errcode));
#endif
		/* FIXME: do something here that lets us return the correct
		 * error here */
		errno = ENETUNREACH;
		retval = -1;
		goto err1;
	}

	max_count = 10;
	if (!(fdlist = calloc(max_count, sizeof(int)))) {
		retval = -1;
		goto err2;
	}

	for (count = 0, aptr = info; aptr != NULL; aptr = aptr->ai_next) {
		if (count > max_count) {
			max_count *= 2;
			if (!(fdlist = realloc(fdlist, max_count
							* sizeof(int)))) {
				retval = -1;
				goto err3;
			}
		}

#ifdef DEBUG
		fprintf(stderr, "Attempting to open socket with family %d socktype %d protocol %d (really 0)\n",
							aptr->ai_family,
							aptr->ai_socktype,
							aptr->ai_protocol);
#endif
		if ((next_fd = exs_socket(aptr->ai_family,
				aptr->ai_socktype, 0)) < 0) {
#ifdef DEBUG
			fprintf(stderr, "Could not open socket with family %d socktype %d protocol %d: %s\n",
							aptr->ai_family,
							aptr->ai_socktype,
							aptr->ai_protocol,
							strerror(errno));
#endif
			continue;
		}

		/* Hack to make Linux listen on IPv4 and IPv6 without
		 * sacrificing portability */
		if (aptr->ai_family == AF_INET6) {
			int yes = 1;
			exs_setsockopt(next_fd, IPPROTO_IPV6, IPV6_V6ONLY, &yes, sizeof(yes));
		}

		if (exs_bind(next_fd, aptr->ai_addr, aptr->ai_addrlen) < 0) {
#ifdef DEBUG
			fprintf(stderr, "Could not bind on fd %d: %s\n",
						next_fd, strerror(errno));
#endif
			exs_blocking_close(next_fd);
			continue;
		}

		if (exs_listen(next_fd, 10) < 0) {
			exs_blocking_close(next_fd);
			continue;
		}

		fdlist[count] = next_fd;
		++count;
	}

	if (count == 0) {
		if (errno == 0) {
			errno = ENODEV;
		}
		retval = -1;
		goto err2;
	}

	*nfds = count;
	retval = 0;
	if (count > 0) {
		*fds = fdlist;
	} else {
		free(fdlist);
	}
	goto err2;

err3:
	for (--count; count >= 0; --count) {
		exs_blocking_close(fdlist[count]);
	}
	free(fdlist);

err2:
	freeaddrinfo(info);

err1:
	return retval;
}	/* our_bind_all */


/* Prints a message containing the given socket address, which can be of any
 * struct sockaddr_* type supported by getnameinfo(3). */
void
print_name_info(int fd, const struct sockaddr *sa, socklen_t salen,
							const char *message)
{
#define NAME_BUFSIZE 100
	char namebuf[NAME_BUFSIZE];
	char servbuf[NAME_BUFSIZE];
	int errcode;

	if ((errcode = getnameinfo(sa, salen, namebuf, NAME_BUFSIZE,
				servbuf, NAME_BUFSIZE,
				NI_NUMERICHOST | NI_NUMERICSERV)) == 0) {
		fprintf(stderr, "fd %d %s %s:%s\n", fd, message, namebuf, servbuf);
	} else {
		fprintf(stderr, "Could not print address for fd %d: %s\n",
				fd, gai_strerror(errcode));
	}
#undef NAME_BUFSIZE
}	/* print_name_info */
