/* libexs/trace.h */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * Parts of this software are copyright Intel Corporation and copied by
 * permission. See the copyright headers of the individual source files
 * for more information. They are provided under the same license.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifndef TRACE_H
#define TRACE_H

void exs_trace_init(void);

void
exs_perror(const char *file, int line, const char *function, int errnum,
		const char *format, ...)
		__attribute__((format(printf, 5, 6)));

void
exs_log(int priority, const char *file, int line, const char *function,
		const char *format, ...)
		__attribute__((format(printf, 5, 6)));

enum exs_log_priority {
	EXS_LOG_ERROR,
	EXS_LOG_INFO,
	EXS_LOG_DEBUG,
};

#define TRACE_ERR(...) \
	exs_log(EXS_LOG_ERROR, \
			__FILE__, __LINE__, \
			__func__, __VA_ARGS__)

#define TRACE_INFO(...) \
	exs_log(EXS_LOG_INFO, \
			__FILE__, __LINE__, \
				__func__, __VA_ARGS__)

#define OUR_PERROR(errcode, ...) \
	exs_perror(__FILE__, __LINE__, __func__, \
			errcode, __VA_ARGS__)

#if defined(ALLOW_EXS_TRACE)

extern int _our_debug_flag;

#define TRACING_IS_ON	_our_debug_flag

#define TRACE_DBG(...) \
	do	{ \
		exs_log(EXS_LOG_DEBUG, __FILE__, __LINE__, \
					__func__, __VA_ARGS__); \
	} while (0)

#define TRACE_VAR_DECLARE(type, name)	type name;

#define TRACE_VAR_ASSIGN(lhs, rhs)	lhs = rhs

#define TRACE_VAR_INCREMENT(var)	var ++

#define MY_NAME(string)						\
	static char	my_name[] = string;

#define PRINT_STATUS(...)					\
	do	{						\
		if(_our_debug_flag) {				\
			our_pstatus(__VA_ARGS__);		\
		}						\
	} while (0)

void
our_dump_buf(FILE *fptr, void *buffer, int length);

#define DUMP_BUF(...)						\
	do	{						\
		if(_our_debug_flag) {				\
			our_dump_buf(stderr, __VA_ARGS__);	\
		}						\
	} while (0)

#define TRACE_EVENT_TYPE(...)					\
	do	{						\
		if(_our_debug_flag) {				\
			our_trace_event_type(__VA_ARGS__);	\
		}						\
	} while (0)						\

/* Wrapper functions for pthread_mutex_lock and pthread_mutex_unlock that print
 * debugging information.  The first argument may be any arbitrary expression
 * that results in a struct or union type.  The second argument MUST be a struct
 * or union member name; it MUST NOT be an arbitrary expression. */
#define OUR_MEMBER_MUTEX_LOCK(ptr, mutex)	do { \
	TRACE_DBG("LOCK " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); \
	pthread_mutex_lock(&(ptr)->mutex); \
	TRACE_DBG("PAST " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); } while (0);

#define OUR_MEMBER_MUTEX_UNLOCK(ptr, mutex)	do { \
	TRACE_DBG("UNLOCK " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); \
	pthread_mutex_unlock(&(ptr)->mutex); } while (0);

#define OUR_MEMBER_RWLOCK_RDLOCK(ptr, rwlock)	do { \
	TRACE_DBG("RDLOCK " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); \
	pthread_rwlock_rdlock(&(ptr)->rwlock); \
	TRACE_DBG("PAST " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); } while (0);

#define OUR_MEMBER_RWLOCK_WRLOCK(ptr, rwlock)	do { \
	TRACE_DBG("WRLOCK " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); \
	pthread_rwlock_wrlock(&(ptr)->rwlock); \
	TRACE_DBG("PAST " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); } while (0);

#define OUR_MEMBER_RWLOCK_UNLOCK(ptr, rwlock)	do { \
	TRACE_DBG("UNLOCK " #ptr " %p " #rwlock " %p", \
		(ptr), &(ptr)->rwlock); \
	pthread_rwlock_unlock(&(ptr)->rwlock); } while (0);

#define OUR_MEMBER_SEM_WAIT(ptr, sem)	do { \
	TRACE_DBG("WAIT " #ptr " %p " #sem " %p", \
		(ptr), &(ptr)->sem); \
	sem_wait(&(ptr)->sem); \
	TRACE_DBG("PAST " #ptr " %p " #sem " %p", \
		(ptr), &(ptr)->sem); } while (0);

#define OUR_MEMBER_SEM_POST(ptr, sem)	do { \
	TRACE_DBG("POST " #ptr " %p " #sem " %p", \
		(ptr), &(ptr)->sem); \
	sem_post(&(ptr)->sem); } while (0);

#define OUR_RWLOCK_RDLOCK(rwlock)	do { \
	TRACE_DBG("RDLOCK " #rwlock " %p", \
		&(rwlock)); \
	pthread_rwlock_rdlock(&(rwlock)); \
	TRACE_DBG("PAST " #rwlock " %p", \
		&(rwlock)); } while (0);

#define OUR_RWLOCK_WRLOCK(rwlock)	do { \
	TRACE_DBG("WRLOCK " #rwlock " %p", \
		&(rwlock)); \
	pthread_rwlock_wrlock(&(rwlock)); \
	TRACE_DBG("PAST " #rwlock " %p", \
		&(rwlock)); } while (0);

#define OUR_RWLOCK_UNLOCK(rwlock)	do { \
	TRACE_DBG("UNLOCK " #rwlock " %p", \
		&(rwlock)); \
	pthread_rwlock_unlock(&(rwlock)); } while (0);

/* TRYLOCK takes a third parameter which is a pointer to an integer in which to
 * place the result from pthread_mutex_trylock. */
#define OUR_MEMBER_MUTEX_TRYLOCK(ptr, mutex, ret)	do { \
	TRACE_DBG("TRYLOCK " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); \
	*(ret) = pthread_mutex_trylock(&(ptr)->mutex); \
	TRACE_DBG("PAST " #ptr " %p " #mutex " %p", \
		(ptr), &(ptr)->mutex); } while (0);

#else

#define TRACING_IS_ON	0

#define TRACE_DBG(module, ...) do { } while (0)

#define TRACE_VAR_DECLARE(type, name)

#define TRACE_VAR_ASSIGN(lhs, rhs)	do { } while (0)

#define TRACE_VAR_INCREMENT(var)	do { } while (0)

#define MY_NAME(string)

#define PRINT_STATUS(args...)					\
	do	{						\
	} while (0)

#define DUMP_BUF(args...)					\
	do	{						\
	} while (0)

#define TRACE_EVENT_TYPE(args...)				\
	do	{						\
	} while (0)

#define OUR_MEMBER_MUTEX_LOCK(ptr, mutex)	do { \
		pthread_mutex_lock(&(ptr)->mutex); \
	} while (0)

#define OUR_MEMBER_MUTEX_UNLOCK(ptr, mutex)	do { \
		pthread_mutex_unlock(&(ptr)->mutex); \
	} while (0)

#define OUR_MEMBER_RWLOCK_RDLOCK(ptr, rwlock)	do { \
		pthread_rwlock_rdlock(&(ptr)->rwlock); \
	} while (0)

#define OUR_MEMBER_RWLOCK_WRLOCK(ptr, rwlock)	do { \
		pthread_rwlock_wrlock(&(ptr)->rwlock); \
	} while (0)

#define OUR_MEMBER_RWLOCK_UNLOCK(ptr, rwlock)	do { \
		pthread_rwlock_unlock(&(ptr)->rwlock); \
	} while (0)

#define OUR_MEMBER_SEM_WAIT(ptr, sem)	do { \
		sem_wait(&(ptr)->sem); \
	} while (0)

#define OUR_MEMBER_SEM_POST(ptr, sem)	do { \
		sem_post(&(ptr)->sem); \
	} while (0)

#define OUR_RWLOCK_RDLOCK(rwlock)	do { \
	pthread_rwlock_rdlock(&(rwlock)); \
	} while (0)

#define OUR_RWLOCK_WRLOCK(rwlock)	do { \
	pthread_rwlock_wrlock(&(rwlock)); \
	} while (0)

#define OUR_RWLOCK_UNLOCK(rwlock)	do { \
	pthread_rwlock_unlock(&(rwlock)); } while (0);

/* TRYLOCK takes a third parameter which is a pointer to an integer in which to
 * place the result from pthread_mutex_trylock. */
#define OUR_MEMBER_MUTEX_TRYLOCK(ptr, mutex, ret)	do { \
	*(ret) = pthread_mutex_trylock(&(ptr)->mutex); } while (0)

#endif

/* This struct contains match rules which are parsed from strings of the form:
 *
 * rule : match_spec
 *
 * match_spec :
 *    'file' word
 *    'line' line_range
 *    'func' word
 *    'function' word
 *    'format' word
 *
 * line_range :
 *    line-line
 *    line-
 *    -line
 *    line
 *
 * word : [^ \n\t\r\f\v]+
 * line : [0-9]+
 *
 * Note that there must be no whitespace in line_range, and that the '-' is a
 * literal dash.  A rule is parsed from a string via our_parse_rule() and
 * each debug print is matched with the entire ruleset in our_match_rules().
 * Rules are loaded from the file provided in the EXS_TRACE_CONFIG environment
 * variable, if we were configured with the appropriate --enable-trace flag.
 */
struct trace_rule {
	char *file;
	int line_min;
	int line_max;
	char *function;
	char *format;
};

#ifdef EXPORT_PARSER
int
our_parse_rule(char *input, char **endptr, struct trace_rule *rule);
#endif

#endif
